// Structure the parameters needed for calculating the likelihood.
// "Parameters" in the sense that they may be subject to optimisation.

use std::fmt::{self, Display};

use arrayvec::ArrayVec;
use color_print::{cwrite, cwriteln};
use serde::{Deserialize, Serialize};
use tch::Tensor;

// The structure is generic among float and tensors
// to allow both likelihood implementations.
// NB: the required bounds to abstract over both in this program
// is very complicated: (f64 - Num, Num / f64, Num.exp(), etc.)
// and it needs to be repeated on every method implementation.
// As a consequence, either duplicate the code or use macros,
// and keep genericity just for factorizing over the parameters type.
pub trait Num: Sized {}
impl Num for f64 {}
impl Num for Tensor {}
impl Num for ValGrad {}

// Constraints to uphold:
//  - all positive
//  - tau_1 <= tau_2
//  - p_* <= 1
//  - p_ab + p_ac + p_bc <= 1
#[derive(Debug, Clone, Serialize, Deserialize, Default)]
pub struct Parameters<F> {
    // Population size.
    pub theta: F,
    // Divergence times.
    pub tau_1: F,
    pub tau_2: F,
    // Gene flow.
    pub p_ab: F,
    pub p_ac: F,
    pub p_bc: F,
    pub p_ancient: F, // Only the most ancient gf time gets a special probability.
    // Date of inferred gene flow events.
    pub gf_times: GeneFlowTimes<F>, // At least one value, sorted decreasing, all positives.
}

// Parameters are stack-allocated, so don't overuse possible GF times.
pub const MAX_N_GF_TIMES: usize = 3;
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Default)]
#[serde(transparent)]
pub struct GeneFlowTimes<F>(pub(crate) ArrayVec<F, MAX_N_GF_TIMES>);

// Convenience bundle.
pub(crate) struct ValGrad {
    pub(crate) value: f64,
    pub(crate) gradient: Option<f64>,
}

//--------------------------------------------------------------------------------------------------
// Minor methods.

impl<F: Num> GeneFlowTimes<F> {
    pub(crate) fn new() -> Self {
        Self(ArrayVec::new())
    }
}

impl<I> GeneFlowTimes<I> {
    pub fn map<O>(&self, f: impl Fn(&I) -> O) -> GeneFlowTimes<O> {
        GeneFlowTimes(self.0.iter().map(f).collect())
    }
}

impl<F: Num> Parameters<F> {
    pub(crate) fn n_gf_times(&self) -> usize {
        self.gf_times.0.len()
    }
}

// Drop gradient information
impl From<&ValGrad> for f64 {
    fn from(val: &ValGrad) -> Self {
        val.value
    }
}

//==================================================================================================
// Display.

impl<F: Num + Display> Display for GeneFlowTimes<F> {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        f.debug_list()
            .entries(self.0.iter().map(|gf| format!("{gf}")))
            .finish()
    }
}

impl Display for ValGrad {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let Self { value, gradient } = self;
        cwrite!(f, "{value:?} <k>'</>")?;
        if let Some(grad) = gradient {
            cwrite!(f, "<k>{grad:?}</>")
        } else {
            cwrite!(f, "<k>{}</>", "<nograd>")
        }
    }
}

impl fmt::Debug for ValGrad {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{self}")
    }
}

// Colored parameters for stdout.
pub struct Colored<'p, F: Num + Display + Sized>(&'p Parameters<F>);

impl<F: Num + Display + Sized + fmt::Debug> Display for Colored<'_, F> {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        writeln!(f, "Parameters {{")?;
        let Parameters {
            theta,
            tau_1,
            tau_2,
            p_ab,
            p_ac,
            p_bc,
            p_ancient,
            gf_times,
        } = &self.0;
        cwriteln!(f, "  theta: <c>{theta:?}</>,")?;
        cwriteln!(f, "  tau_1: <c>{tau_1:?}</>,")?;
        cwriteln!(f, "  tau_2: <c>{tau_2:?}</>,")?;
        cwriteln!(f, "  p_ab: <c>{p_ab:?}</>,")?;
        cwriteln!(f, "  p_ac: <c>{p_ac:?}</>,")?;
        cwriteln!(f, "  p_bc: <c>{p_bc:?}</>,")?;
        cwriteln!(f, "  p_ancient: <c>{p_ancient:?}</>,")?;
        writeln!(f, "  gf_times: [")?;
        for t in &gf_times.0 {
            cwriteln!(f, "    <c>{t}</>,")?;
        }
        writeln!(f, "  ]")?;
        writeln!(f, "}}")
    }
}

impl<F: Num + Display + Sized> Parameters<F> {
    pub fn colored(&self) -> Colored<F> {
        Colored(self)
    }
}
